from typing import Optional

import tensorflow as tf
from tensorflow import keras
from tensorflow.keras.metrics import AUC

from sfu_ml_lib.base.metrics import Metric


class PredictionBias(keras.metrics.Metric):
    def __init__(self) -> None:
        super().__init__(name=Metric.PREDICTION_BIAS.value, dtype=tf.float32)

        self.target_mean = self.add_weight('target_mean', initializer=tf.zeros_initializer)
        self.prediction_mean = self.add_weight('prediction_mean', initializer=tf.zeros_initializer)
        self.count = self.add_weight('count', initializer=tf.zeros_initializer)

    @staticmethod
    def update_mean(mean: tf.Variable, count: tf.Variable, batch_size: tf.Tensor, values: tf.Tensor) -> None:
        new_count = count + batch_size
        new_mean = (tf.reduce_sum(values) + count * mean) / new_count

        mean.assign(new_mean)

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        assert sample_weight is None

        batch_size = tf.cast(tf.shape(y_true)[0], dtype=tf.float32)
        predictions = tf.round(y_pred)

        self.update_mean(self.target_mean, self.count, batch_size, y_true)
        self.update_mean(self.prediction_mean, self.count, batch_size, predictions)

        self.count.assign_add(batch_size)

    def result(self) -> tf.Tensor:
        return self.prediction_mean - self.target_mean

    def reset_state(self) -> None:
        self.target_mean.assign(tf.zeros(()))
        self.prediction_mean.assign(tf.zeros(()))
        self.count.assign(tf.zeros(()))


class CutoffThreshold(keras.metrics.Metric):
    def __init__(self) -> None:
        super().__init__(name=Metric.CUTOFF_THRESHOLD.value, dtype=tf.float32)

        self.auc = AUC()

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        self.auc.update_state(y_true, y_pred, sample_weight)

    def result(self) -> tf.Tensor:
        tpr = tf.math.divide_no_nan(self.auc.true_positives, self.auc.true_positives + self.auc.false_negatives)
        fpr = tf.math.divide_no_nan(self.auc.false_positives, self.auc.false_positives + self.auc.true_negatives)

        index = tf.argmax(tpr - fpr)
        threshold = tf.convert_to_tensor(self.auc.thresholds)[index]

        return threshold

    def reset_states(self) -> None:
        self.auc.reset_states()



import json
import os
from typing import Optional, Mapping

import sfu_data_io.helpers as io
from tensorflow.keras.callbacks import Callback

from sfu_ml_lib.base.tracker import Tracker
from sfu_ml_lib.base.metrics import NumericType


class TrackerCallback(Callback):
    def __init__(self, tracker: Tracker, split: Optional[int] = None, save: bool = False) -> None:
        super().__init__()

        self.tracker = tracker
        self.split = split
        self.save = save

    @classmethod
    def to_string(cls, metrics: Mapping[str, NumericType], prefix: Optional[str] = None) -> str:
        string_metrics = [
            f'{key}: {value}'
            for key, value
            in sorted(metrics.items(), key=lambda key_value: key_value[0])
        ]

        string_metrics = [prefix] + string_metrics if prefix is not None else string_metrics

        message = ', '.join(string_metrics)

        return message

    def save_model(self) -> None:
        local_prefix = os.path.join(io.get_temporary_local_directory_path(), 'weights')

        self.model.save_weights(local_prefix)
        self.tracker.save_model(local_prefix)

    def save_config(self) -> None:
        local_file = os.path.join(io.get_temporary_local_directory_path(), 'config.json')

        with io.open(local_file, mode='w') as json_file:
            json.dump(self.model.get_config(), json_file)

        self.tracker.save_model(local_file)

    def on_epoch_end(self, epoch: int, logs: Optional[Mapping[str, NumericType]] = None):
        if logs is not None:
            self.tracker.log_metrics(logs, step=epoch)
            print(self.to_string(logs, f'Finished Epoch: {epoch}'))

    def on_train_end(self, logs: Optional[Mapping[str, NumericType]] = None):
        if self.save and self.model is not None:
            self.save_model()

            try:
                self.save_config()
            except NotImplementedError:
                pass

        if logs is not None:
            self.tracker.log_metrics(logs, step=self.split)
            print(self.to_string(logs, f'Finished Split: {self.split}'))



import typing
from abc import ABC, abstractmethod
from typing import Tuple, Optional, Iterator, Iterable, TypeVar, Union

import numpy as np
import tensorflow as tf
from tensorflow import Tensor, TensorSpec, RaggedTensorSpec, SparseTensorSpec


T = TypeVar('T')

TensorType = Union[np.ndarray, Tensor]
TensorTuple = Union[TensorType, Tuple[TensorType, ...]]
BatchType = Union[
    Tuple[TensorTuple, None],
    Tuple[TensorTuple, TensorTuple],
    Tuple[TensorTuple, TensorTuple, TensorType],
]

Specification = Union[SparseTensorSpec, RaggedTensorSpec, TensorSpec]
Schema = Union[Specification, Iterable['Schema']]  # type: ignore


class Dataset(ABC):
    input_schema: Schema
    size: int
    targets: Optional[TensorType] = None

    @abstractmethod
    @typing.no_type_check
    def get_batches(self, *args) -> tf.data.Dataset:
        ...

    @abstractmethod
    def subset(self: T, indices: TensorType) -> T:
        ...

    def create_tf_dataset(self, iterator: Iterator[BatchType], queue_size: int = -1) -> tf.data.Dataset:
        return tf.data.Dataset.from_generator(iterator, output_signature=self.input_schema).prefetch(queue_size)



import math
from typing import Tuple, Generator

import numba
import numpy as np


@numba.njit
def compute_target_rates(targets: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes the counts and rates of a target vector.

    >>> compute_target_rates(np.array([0, 0, 0, 0, 1]))
    (array([1., 4.], dtype=float32), array([4, 1], dtype=int32))

    :param targets: A Numpy vector of target indices.
    :return: A tuple with rates and counts.
    """
    num_classes = np.max(targets) + 1

    counts = np.zeros(num_classes, dtype=np.int32)
    for target in targets:
        counts[target] += 1

    rates = (np.max(counts) / counts).astype(np.float32)

    return rates, counts


@numba.njit
def get_oversampled_indices(targets: np.ndarray) -> np.ndarray:
    rates, counts = compute_target_rates(targets)

    quotients = np.array([math.floor(value) for value in rates], dtype=np.int32)
    remainders = [round(value) for value in (rates - quotients) * counts]

    indexed_targets = np.array(list(enumerate(targets)), dtype=targets.dtype)
    np.random.shuffle(indexed_targets)

    indices = []

    for index, target in indexed_targets:
        for _ in range(quotients[target]):
            indices.append(index)

        if remainders[target] > 0:
            indices.append(index)
            remainders[target] -= 1

    indices = np.array(indices, np.int32)

    return indices


@numba.njit
def get_undersampled_indices(targets: np.ndarray, batch_size: int) -> Generator[np.ndarray, None, None]:
    num_classes = np.max(targets) + 1

    class_indices = [
        numba.typed.List.empty_list(numba.types.int64)
        for _ in range(num_classes)
    ]

    for index, target in enumerate(targets):
        class_indices[target].append(index)

    while True:
        class_size = min(math.floor(batch_size / num_classes), min([len(indices) for indices in class_indices]))

        if class_size == 0:
            break

        batch_indices = []

        for _ in range(class_size):
            for indices in class_indices:
                batch_indices.append(indices.pop())

        yield np.array(batch_indices, dtype=np.int32)


def get_oversampled_indices_subset(indices: np.ndarray, targets: np.ndarray, balance: bool = True) -> np.ndarray:
    if not balance:
        return indices

    oversampled_indices = get_oversampled_indices(targets[indices])
    indices = indices[oversampled_indices]

    return indices


def get_undersampled_indices_subset(
        indices: np.ndarray,
        targets: np.ndarray,
        batch_size: int,
        balance: bool) -> np.ndarray:

    if balance:
        for batch_indices in get_undersampled_indices(targets[indices], batch_size):
            yield indices[batch_indices]

    else:
        for start_index in range(0, indices.size, batch_size):
            yield indices[start_index:start_index + batch_size]



from typing import Optional

import tensorflow as tf
from tensorflow import keras

from sfu_ml_lib.base.metrics import Metric


class PearsonCorrelation(keras.metrics.Metric):
    def __init__(self) -> None:
        super().__init__(name=Metric.PEARSON_CORRELATION.value, dtype=tf.float32)

        self.target_mean = self.add_weight('target_mean', initializer=tf.zeros_initializer)
        self.target_square_mean = self.add_weight('target_square_mean', initializer=tf.zeros_initializer)
        self.prediction_mean = self.add_weight('prediction_mean', initializer=tf.zeros_initializer)
        self.prediction_square_mean = self.add_weight('prediction_square_mean', initializer=tf.zeros_initializer)
        self.product_mean = self.add_weight('product_mean', initializer=tf.zeros_initializer)
        self.count = self.add_weight('count', initializer=tf.zeros_initializer)

    @classmethod
    def update_mean(cls, mean: tf.Variable, count: tf.Variable, values: tf.Tensor) -> None:
        new_count = count + tf.cast(tf.shape(values)[0], dtype=tf.float32)
        new_mean = (tf.reduce_sum(values) + count * mean) / new_count

        mean.assign(new_mean)

    def update_state(self, y_true: tf.Tensor, y_pred: tf.Tensor, sample_weight: Optional[tf.Tensor] = None) -> None:
        self.update_mean(self.target_mean, self.count, y_true)
        self.update_mean(self.target_square_mean, self.count, tf.square(y_true))
        self.update_mean(self.prediction_mean, self.count, y_pred)
        self.update_mean(self.prediction_square_mean, self.count, tf.square(y_pred))
        self.update_mean(self.product_mean, self.count, y_true * y_pred)

        self.count.assign_add(tf.cast(tf.shape(y_true)[0], dtype=tf.float32))

    def result(self) -> tf.Tensor:
        covariance = (self.product_mean - self.target_mean * self.prediction_mean)
        target_standard_deviation = tf.sqrt(self.target_square_mean - tf.square(self.target_mean))
        prediction_standard_deviation = tf.sqrt(self.prediction_square_mean - tf.square(self.prediction_mean))

        pearson_correlation = covariance / (target_standard_deviation * prediction_standard_deviation)

        return pearson_correlation

    def reset_state(self) -> None:
        self.target_mean.assign(tf.zeros(()))
        self.target_square_mean.assign(tf.zeros(()))
        self.prediction_mean.assign(tf.zeros(()))
        self.prediction_square_mean.assign(tf.zeros(()))
        self.product_mean.assign(tf.zeros(()))
        self.count.assign(tf.zeros(()))



import random
from typing import List, Generator

import numpy as np


def get_batches(
        data: List[np.ndarray],
        batch_size: int,
        pad_batch: bool = False) -> Generator[List[np.ndarray], None, None]:
    """
    Get a batch iterator.
    :param data: List of data you like to create batches from. The first dimension
    of the array represents number of data points.
    :param batch_size: The size of each batch.
    :param pad_batch: True if all batches need to be the same size
    """
    data_size = len(data)

    if pad_batch:
        delta = data_size - batch_size * int(data_size / batch_size)
        data.extend(random.sample(data, delta))

    for start_index in range(0, data_size, batch_size):
        yield data[start_index:start_index + batch_size]



import traceback
from contextlib import contextmanager
from typing import Mapping, Optional

from sfu_ml_lib.base.metrics import NumericType, PrimitiveType


class Tracker:
    def set_tags(self, tags: Mapping[str, str]) -> None:
        raise NotImplementedError

    def register_parameters(self, parameters: Mapping[str, PrimitiveType]) -> None:
        raise NotImplementedError

    def log_metrics(self, metrics: Mapping[str, NumericType], step: Optional[int] = None) -> None:
        raise NotImplementedError

    def save_model(self, path: str) -> None:
        raise NotImplementedError

    def finish_epoch(self) -> None:
        raise NotImplementedError

    def __enter__(self) -> 'Tracker':
        raise NotImplementedError

    def __exit__(self, context_type, value, trace_back) -> None:
        raise NotImplementedError

    @contextmanager
    def exception_logging(self, limit: int = 10):
        try:
            yield

        except Exception as error:
            self.set_tags({'Error': traceback.format_exc(limit=limit)})
            raise error



from typing import Tuple, Type

import numba
import numpy as np


@numba.njit()
def convert_incremental_lengths_to_indices(accumulated_lengths: np.ndarray) -> np.ndarray:
    """
    Convert a vector of accumulated item lengths into a matrix of indices.
    For instance, given [0, 2, 5], we obtain [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]].
    :param accumulated_lengths: A vector of accumulated lengths of size N + 1.
    :return: A matrix of size (total length, 2) representing the assignment of each index to an item in N.
    """
    num_indices = accumulated_lengths[-1] - accumulated_lengths[0]

    indices = np.empty((num_indices, 2), dtype=np.int32)

    index = 0

    for outer_index in range(accumulated_lengths.size - 1):
        length = accumulated_lengths[outer_index + 1] - accumulated_lengths[outer_index]

        for inner_index in range(length):
            indices[index, 0] = outer_index
            indices[index, 1] = inner_index

            index += 1

    return indices


@numba.njit()
def convert_lengths_to_indices(lengths: np.ndarray) -> np.ndarray:
    """
    Convert a vector of item lengths into a matrix of indices.
    For instance, given [2, 3], we obtain [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]].
    :param lengths: A vector of lengths of size N.
    :return: A matrix of size (total length, 2) representing the assignment of each index to an item in N.
    """
    num_indices = np.sum(lengths)

    indices = np.empty((num_indices, 2), dtype=np.int32)

    index = 0

    for outer_index, length in enumerate(lengths):
        for inner_index in range(length):
            indices[index, 0] = outer_index
            indices[index, 1] = inner_index

            index += 1

    return indices


@numba.njit()
def convert_segment_ids_to_indices(segment_ids: np.ndarray) -> np.ndarray:
    """
    Convert a vector of segment ids into a matrix of indices.
    For instance, given [0, 0, 1, 1, 1], we obtain [[0, 0], [0, 1], [1, 0], [1, 1], [1, 2]].
    :param segment_ids: A vector of segment ids of size N.
    :return: A matrix of size (N, 2) representing the assignment of each index to an item in N.
    """
    indices = np.empty((segment_ids.size, 2), dtype=np.int32)

    previous_outer_index, inner_index = segment_ids[0], 0

    for index, outer_index in enumerate(segment_ids):
        if previous_outer_index != outer_index:
            inner_index = 0

        indices[index, 0] = outer_index
        indices[index, 1] = inner_index

        inner_index += 1
        previous_outer_index = outer_index

    return indices


@numba.njit()
def convert_incremental_lengths_to_segment_indices(accumulated_lengths: np.ndarray) -> np.ndarray:
    """
    Convert a vector of accumulated item lengths into a segment of indices.
    For instance, given [0, 2, 5], we obtain [0, 0, 1, 1, 1].
    :param accumulated_lengths: A vector of accumulated lengths of size N + 1.
    :return: A vector of size (total length) representing the assignment of each item index in N.
    """
    num_indices = accumulated_lengths[-1] - accumulated_lengths[0]

    segment_indices = np.empty(num_indices, dtype=np.int32)

    index = 0

    for outer_index in range(accumulated_lengths.size - 1):
        length = accumulated_lengths[outer_index + 1] - accumulated_lengths[outer_index]

        for _ in range(length):
            segment_indices[index] = outer_index

            index += 1

    return segment_indices


@numba.njit()
def convert_incremental_lengths_to_lengths(accumulated_lengths: np.ndarray) -> np.ndarray:
    """
    Convert a vector of accumulated item lengths into lengths.
    For instance, given [0, 2, 5], we obtain [2, 3].
    :param accumulated_lengths: A vector of accumulated lengths of size N + 1.
    :return: A vector of size N representing the length of each item.
    """
    num_items = accumulated_lengths.size - 1

    lengths = np.empty(num_items, dtype=np.int32)

    for index in range(num_items):
        lengths[index] = accumulated_lengths[index + 1] - accumulated_lengths[index]

    return lengths


@numba.njit()
def accumulate(vector: np.ndarray) -> np.ndarray:
    """
    Accumulates the values of a vector such that accumulated[i] = accumulated[i - 1] + vector[i]
    :param vector: A vector of size N.
    :return: A vector of size N + 1.
    """
    accumulated = np.zeros(vector.size + 1, dtype=np.int32)

    for index, value in enumerate(vector):
        accumulated[index + 1] = accumulated[index] + value

    return accumulated


@numba.njit()
def get_sparse_coo_subset(
        indptr: np.ndarray,
        indices: np.ndarray,
        data: np.ndarray,
        row_indices: np.ndarray,
        dtype: Type) -> Tuple[np.ndarray, np.ndarray]:
    """
    Given a sparse matrix in CRS format, we return a subset in COO format.
    :param indptr: The cumulative sum of non-zero values per row.
    :param indices: The column indices for each row.
    :param data: The non-zero values going throw each column first.
    :param row_indices: The row indices to subset by.
    :param dtype: The type of the indices.
    :return: A tuple with a (N, 2) matrix specifying matrix coordinates and a vector with non-zero values.
    """
    sparse_indices, data_subset = [], []

    for batch_index, row_index in enumerate(row_indices):
        for sparse_index in range(indptr[row_index], indptr[row_index + 1]):
            sparse_indices.append((batch_index, indices[sparse_index]))
            data_subset.append(data[sparse_index])

    sparse_indices = np.array(sparse_indices, dtype=dtype)
    data_subset = np.array(data_subset, dtype=data.dtype)

    return sparse_indices, data_subset


@numba.njit()
def get_sparse_subset(
        indptr: np.ndarray,
        indices: np.ndarray,
        data: np.ndarray,
        row_indices: np.ndarray,
        dtype: Type) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Given a sparse matrix in CRS format, we return a subset in the same format.
    :param indptr: The cumulative sum of non-zero values per row.
    :param indices: The column indices for each row.
    :param data: The non-zero values going throw each column first.
    :param row_indices: The row indices to subset by.
    :param dtype: The type of the indices.
    :return: A tuple with a subset of `indptr`, `indices` and `data`.
    """
    indptr_subset, indices_subset, data_subset = [0], [], []

    for row_index in row_indices:
        num_indices = indptr[row_index + 1] - indptr[row_index]

        indptr_subset.append(indptr_subset[-1] + num_indices)

        for sparse_index in range(indptr[row_index], indptr[row_index + 1]):
            indices_subset.append(indices[sparse_index])
            data_subset.append(data[sparse_index])

    indptr_subset = np.array(indptr_subset, dtype=dtype)
    indices_subset = np.array(indices_subset, dtype=dtype)
    data_subset = np.array(data_subset, dtype=data.dtype)

    return indptr_subset, indices_subset, data_subset



import glob
from typing import Mapping, Collection, Optional

import mlflow

from sfu_ml_lib.base.metrics import NumericType, PrimitiveType
from sfu_ml_lib.base.tracker import Tracker
from sfu_ml_lib.metrics.aggregator import Aggregation, MetricAggregator


class MLFlowTracker(MetricAggregator, Tracker):
    def __init__(
            self,
            experiment_name: str,
            tracking_uri: Optional[str] = None,
            aggregators: Optional[Collection[Aggregation]] = None):

        super().__init__(aggregators)

        self.step = 0

        if tracking_uri:
            mlflow.set_tracking_uri(tracking_uri)

        mlflow.set_experiment(experiment_name)

    def set_tags(self, tags: Mapping[str, str]) -> None:
        mlflow.set_tags(tags)

    def register_parameters(self, parameters: Mapping[str, PrimitiveType]) -> None:
        mlflow.log_params({self.to_str(key): value for key, value in parameters.items()})

    def log_metrics(self, metrics: Mapping[str, NumericType], step: Optional[int] = None) -> None:
        mlflow.log_metrics(metrics, step)

    def save_model(self, path: str) -> None:
        for file_path in glob.glob(path + '*'):
            mlflow.log_artifact(file_path)

    def finish_epoch(self) -> None:
        metrics = {self.to_str(key): value for key, value in self.get_metrics().items()}

        self.log_metrics(metrics, self.step)

        self.flush()

        self.step += 1

    def __enter__(self) -> 'MLFlowTracker':
        return self

    def __exit__(self, context_type, value, traceback) -> None:
        mlflow.end_run()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sfu_ml_lib',
 'sfu_ml_lib.base',
 'sfu_ml_lib.common',
 'sfu_ml_lib.dataset',
 'sfu_ml_lib.metrics',
 'sfu_ml_lib.metrics.aggregation',
 'sfu_ml_lib.models',
 'sfu_ml_lib.modules',
 'sfu_ml_lib.trackers']

package_data = \
{'': ['*']}

install_requires = \
['mlflow>=1.8,<2.0',
 'numba>=0.51,<1',
 'numpy',
 'scikit-learn>=0.24,<1',
 'scipy>=1.4,<2.0',
 'sfu-data-io>=0.0,<0.1',
 'tensorflow>=2.2,<3.0']

setup_kwargs = {
    'name': 'sfu-ml-lib',
    'version': '0.0.7',
    'description': 'Libraries that support the development of machine learning models in TensorFlow.',
    'long_description': None,
    'author': 'Anderson de Andrade',
    'author_email': 'adeandrade@ensc.sfu.ca',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from enum import Enum
from typing import Mapping, Union, MutableMapping, Callable

import numpy as np
import tensorflow as tf


NumericType = Union[int, float]
PrimitiveType = Union[int, float, str, bool]
AggregatorType = Callable[..., NumericType]
MetricKeyType = Union[str, 'Metric']
MetricValueType = Union[NumericType, np.ndarray, tf.Tensor]


TRAIN_PREFIX = 'Training'
VALIDATION_PREFIX = 'Validation'
TEST_PREFIX = 'Test'


class Metric(Enum):
    LOSS = 'Loss'

    ACCURACY = 'Accuracy'
    ROC_AUC = 'ROC AUC'
    PREDICTION_BIAS = 'Prediction Bias'
    CUTOFF_THRESHOLD = 'Cutoff Threshold'

    COSINE_SIMILARITY = 'Cosine Similarity'
    COSINE_DISTANCE = 'Cosine Distance'
    JACCARD_SCORE = 'Jaccard Score'
    CROSS_ENTROPY = 'Cross Entropy'
    MUTUAL_INFORMATION = 'Mutual Information'
    REWARD = 'Reward'

    NOISE_CONTRASTIVE_ESTIMATION = 'Noise Contrastive Estimation'
    NORMALIZED_TEMPERATURE_SCALED_CROSS_ENTROPY = 'Normalized Temperature-Scaled Cross Entropy'

    MEAN_ABSOLUTE_ERROR = 'Mean Absolute Error'
    MEAN_SQUARE_ERROR = 'Mean Square Error'
    PEARSON_CORRELATION = 'Pearson Correlation'

    WEIGHTED_F1 = 'Weighted F1 Score'
    WEIGHTED_PRECISION = 'Weighted Precision'
    WEIGHTED_RECALL = 'Weighted Recall'

    MICRO_F1 = 'Micro F1 Score'
    MICRO_PRECISION = 'Micro Precision'
    MICRO_RECALL = 'Micro Recall'

    MACRO_F1 = 'Macro F1 Score'
    MACRO_PRECISION = 'Macro Precision'
    MACRO_RECALL = 'Macro Recall'

    @property
    def train_value(self) -> str:
        return f'{TRAIN_PREFIX} {self.value}'

    @property
    def validation_value(self) -> str:
        return f'{VALIDATION_PREFIX} {self.value}'

    @property
    def test_value(self) -> str:
        return f'{TEST_PREFIX} {self.value}'


def prefix_metrics_train(metrics: Mapping[MetricKeyType, MetricValueType]) -> MutableMapping[str, NumericType]:
    return prefix_metrics(metrics, TRAIN_PREFIX)


def prefix_metrics_validation(metrics: Mapping[MetricKeyType, MetricValueType]) -> MutableMapping[str, NumericType]:
    return prefix_metrics(metrics, VALIDATION_PREFIX)


def prefix_metrics_test(metrics: Mapping[MetricKeyType, MetricValueType]) -> MutableMapping[str, NumericType]:
    return prefix_metrics(metrics, TEST_PREFIX)


def to_string(key: MetricKeyType) -> str:
    return key.value if isinstance(key, Enum) else key


def prefix_metrics(metrics: Mapping[MetricKeyType, MetricValueType], prefix: str) -> MutableMapping[str, NumericType]:
    return {f'{prefix} {to_string(key)}': value for key, value in metrics.items()}
